# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# -------------------------------------------------------------------
#  $RCSfile: BARP.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:43 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package BARP;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(read innerlay_ids innerlay_ips);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use Net::IP;

# Globals
#
my %barp;                # BGP/ARP table (initialized from $barpfile)


# Purpose:
#       Return a list of innerlay IDs.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       -
#
# Returns:
#       list of innerlay IDs
#
# Exceptions:
#       thrown on error
#
sub innerlay_ids () {
  my @ips = keys %barp;
  return keys %{ $barp{pop @ips} };
}


# Purpose:
#       Read in the BARP table from a file.
#
# Arguments:
#       $barpfile        Data::Dumper file containing the BARP table
#
#       $outerlay_addr   outerlay IP address of this host (Net::IP object)
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub read ($$) {
  my ($barpfile, $outerlay_addr) = @_;

  # read in the BARP table
  open FH, $barpfile or die "Cannot open $barpfile: $!";
  my $barptext = "";
  while (<FH>) { chomp; $barptext .= $_; }
  close FH or die "Cannot close $barpfile: $!";
  
  # untaint BARP file contents and evaluate the contents
  # XXX security hole!
  if ($barptext =~ /(.*)/) { $barptext = $1; }
  else { usage "Invalid BARP table contents"; }
  eval $barptext or die "Error evaluting BARP table: $@";
  
  # sanity check: BARP table must contain local outerlay IP address
  usage "BARP table does not contain outerlay address $outerlay_addr->ip"
    unless exists $barp{$outerlay_addr->ip};
  
  # sanity check: innerlay IDs in all BARP entries must be identical
  foreach my $o (keys %barp) {
    usage "some BARP entries have different innerlay IDs"
      if different innerlay_ids, keys %{ $barp{$o} }
    }
  
  # sanity check: all innerlay IPs must be different
  my %ips;
  foreach my $o (keys %barp) {
    foreach my $i (innerlay_ids) {
      usage "Non-unique innerlay IP $barp{$o}{$i} in BARP"
	if exists $ips{$barp{$o}{$i}};
      $ips{$barp{$o}{$i}} = 1;

      # make a Net::IP object from the textual IP address
      $barp{$o}{$i} = new Net::IP $barp{$o}{$i} or die Net::IP::Error;
    }
  }
}


# Purpose:
#       Return the hash of innerlay IP addresses a PRM has on
#       different innerlays.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $ip    IP address of the PRM whose innerlay IP addresses should be
#              returned
#
# Returns:
#       hash ID->IP, where ID is an innerlay ID, and IP a Net::IP object
#
# Exceptions:
#       thrown on error
#
sub innerlay_ips ($) {
  my ($ip) = @_;
  $ip = $ip->ip;
  sysdie "no entry in BARP table for $ip" unless exists $barp{$ip};
  return %{ $barp{$ip} };
}


1;
