# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# -------------------------------------------------------------------
#  $RCSfile: Tun.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package Tun;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(create destroy);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use POSIX;
use IO::File;


# Constants
#
use constant IFCONFIG  => "/sbin/ifconfig "; # path to ifconfig command



# Purpose:
#       Configure tun to capture outgoing outerlay packets.
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $outerlay_addr    outerlay IP address (Net::IP object)
#
#       $outerlay_block   outerlay block (Net::IP object)
#
# Returns:
#       ($fh, $tundev)    pair of a non-blocking, read/write file handle to
#                         the configured tun device, and its name
#       device
#
# Exceptions:
#       thrown on error
#
sub create ($$) {
  my ($outerlay_addr, $outerlay_block) = @_;
  
  # Find an available tun device.
  my $i = 0;
  my $cmd = IFCONFIG ."-lu";
  foreach my $if (split / /, `$cmd`) {
    $i++ if $if =~ /tun$i/;
  }
  my $tundev = "tun$i";

  # open the tun device read/write
  my $fh = new IO::File;
  $fh->open("+>/dev/$tundev");
  defined $fh or sysdie "Cannot open /dev/$tundev: $!";
  
  # make it broadcast (default is point-to-point)
  $i = pack "i", 0x2;           # 0x2        is IFF_BROADCAST
  $fh->ioctl(0x8004745e, $i)    # 0x8004745e is TUNSIFMODE
    or sysdie "Cannot ioctl $tundev, TUNSIFMODE, IFF_BROADCAST: $!";
  sysprint "ioctl $tundev, TUNSIFMODE, IFF_BROADCAST" if $verbose;
  
  # set the IP address/netmask
  # XXX setting the MTU is a horrible hack, our IPIP must support PMTU instead
  $cmd = join " ", (IFCONFIG, $tundev, $outerlay_addr->ip, 
		    "netmask", $outerlay_block->mask, "up", "mtu", 1200);
  sysprint $cmd if $verbose;
  system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";
  
  # make it non-blocking, otherwise later read calls will block
  $fh->autoflush(1);
  fcntl $fh, F_SETFL, O_NONBLOCK 
    or sysdie "Cannot fcntl $tundev, F_SEFTL, O_NONBLOCK: $!";
  sysprint "fcntl $tundev, F_SEFTL, O_NONBLOCK" if $verbose;

  # return the file handle (unless exceptions occurred)
  return ($fh, $tundev) unless $@;
}



# Purpose:
#       Destroy tun device (= undo tun_create operation).
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh       file handle of the tun device (return value from
#                 tun_create)
#       $tundev   name of the tun device
#
# Returns:
#       1
#
# Exceptions:
#       -
#
sub destroy ($$) {
  my ($fh, $tundev) = @_;

  # bring interface down and delete its configuration
  foreach my $operation (qw(down delete)) {
    my $cmd = IFCONFIG. "$tundev $operation";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or syswarn "Cannot execute $cmd: $!";
  }

  if (defined $fh) {
    # make interface point-to-point again
    my $i = pack "i", 0x10;     # 0x10       is IFF_POINTOPOINT
    $fh->ioctl(0x8004745e, $i)  # 0x8004745e is TUNSIFMODE
      or syswarn "Cannot ioctl $tundev, TUNSIFMODE, IFF_POINTOPOINT: $!";
    sysprint "ioctl $tundev, TUNSIFMODE, IFF_POINTOPOINT" if $verbose;
    
    # close file handle
    $fh->close;
  }
  
  return 1;
}


1;
