# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# ----------------------------------------------------------------------------
#  $RCSfile: Utilities.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package Utilities;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(usage syswarn sysdie sysprint daemon different 
	     ip_block sysctl_read);
@EXPORT_OK = qw($debug $verbose);

use strict;
use English;
use POSIX;
use Net::IP qw(:PROC);
use Sys::Syslog;


# Constants
#
use constant SYSCTL    => "/sbin/sysctl";    # path to sysctl command


# Globals
#
our $debug = 0;
our $verbose = 0;


# Purpose:
#       Write error/warning/info message to syslog or stdout, depending
#       on the value of $debug. Syswarn and sysprint return to the
#       caller, sysdie aborts the program.
#
# Arguments:
#       $msg   message to write
#
# Returns:
#       -      (syswarn aborts the program)
#
# Exceptions:
#       -
#

BEGIN {
  # open syslog
  Sys::Syslog::setlogsock "unix" or die "Cannot set AF_UNIX syslog: $!";
  openlog $FindBin::Script, "cons,pid", "user" or die "Cannot open syslog: $!";
}

END {
  closelog;
}

sub syswarn ($;$) {
  my ($msg, $level) = @_;
  $level = "warning" unless defined $level;
  if ($debug) { chomp $msg; print "$msg\n"; }
  else { syslog $level, $msg; }
}
	     
sub sysdie ($) { syswarn shift, "err"; kill TERM => $PID; }
sub sysprint ($) { syswarn shift, "info"; }



# Purpose:
#       Become a daemon. (This code was adapted from Proc::Daemon.)
#
# Arguments:
#       -
#
# Returns:
#       -
#
# Exceptions:
#       -
#
sub daemon () {
  sub _fork {
    if (defined (my $pid = fork)) { return $pid; }
    else { sysdie "Can't fork: $!"; }
  }

  # fork and exit parent
  if (_fork) { exit 0; }

  # detach ourselves from the terminal
  POSIX::setsid or sysdie "Cannot detach from controlling terminal";

  # prevent possibility of acquiring a controlling terminal
  $SIG{'HUP'} = 'IGNORE';
  if (_fork) { exit 0; }
  
  # change working directory
  chdir "/";
  
  # clear file creation mask
  umask 0;
  
  # close open file descriptors
  foreach my $i (0..POSIX::sysconf(&POSIX::_SC_OPEN_MAX)) { POSIX::close $i; }
  
  # reopen stderr, stdout, stdin to /dev/null
  open STDIN,  "+>/dev/null";
  open STDOUT, "+>&STDIN";
  open STDERR, "+>&STDIN";
}



# Purpose:
#       Determine whether two arrays are different (do not contain
#       all identical entries).
#
# Arguments:
#       @array1   first array to compare
#       @array2   second array to compare
#
# Returns:
#       0         if the arrays are identical
#
# Exceptions:
#       -
#
sub different (@@) {
  # Stolen from perlfaq4 question "How do I compute the difference of
  # two arrays? How do I compute the intersection of two arrays?"
  # The original code is actually much more powerful, but has been
  # reduced to a simple difference test.

  my (@a1, @a2) = @_;
  my (@inter, @diff) = ((), ());
  my %c = ();
  
  foreach my $i (@a1, @a2) { $c{$i}++ }
  foreach my $i (keys %c) { push @{ $c{$i} > 1 ? \@inter : \@diff }, $i; }
  return scalar @diff;
}


# Purpose:
#       Translate a string IP_ADDRESS/MASK into a string IP_BLOCK/MASK,
#       where the difference between IP_ADDRESS and IP_BLOCK is that the
#       bits outside the subnet mask are zero in IP_BLOCK. This function
#       is mostly useful when creating Net::IP objects.
#
#       Example: 10.26.8.11/24 -> 10.26.8.0/24
#
# Arguments:
#       $ip   string to translatefirst array to compare
#
# Returns:
#       Translated string.
#
# Exceptions:
#       -
#
sub ip_block ($) {
  my ($ip, $mask) = split "/", shift;
  my $bin_ip = ip_iptobin $ip, 4;
  $bin_ip = substr $bin_ip, 0, $mask;
  $bin_ip .= "0" x (32 - $mask);
  $ip = ip_bintoip $bin_ip,4 ;
  return "$ip/$mask";
}


# Description: 
#       Return value of sysctl variable $var.
#
# Arguments:
#       $var    sysctl variable to check
#
# Returns:
#       value of $var if found
#       undef if $var not found
#
# Exceptions:
#       thrown on failure
#
sub sysctl_read ($) {
  my $var = shift; # variable to read
  my $val;
  
  # get the value by reading from a pipe to sysctl
  my $pipe = SYSCTL ." -n $var 2> /dev/null |";
  open PIPE, $pipe or sysdie "Cannot open pipe $pipe: $!";
  # read the output
  $val = <PIPE>;
  # done with pipe
  close PIPE;
  # post-process
  if(defined $val) { chomp $val; }
  return $val;
}


# Purpose:
#       Print info about command line arguments to stdout and
#       terminate.
#
# Arguments:
#       $msg    error message (omitted or empty -> no error message)
#
# Returns:
#       -
#
# Exceptions:
#       -
#
sub usage ($) {
  my $msg = shift;

  print "ERROR:    $msg\n" if defined $msg and $msg ne "";
  print <<EOB

USAGE:    $FindBin::Script -outerlay <IP/mask> -barp <file>


REQUIRED: -outerlay    IP address and netmask of the outerlay
                       (A.B.C.D/E notation)

          -barp        file name of the BGP/ARP table (Data::Dumper format)


OPTIONAL: -verbose     verbose mode

          -debug       do not become a daemon, and log to stdout not syslog

          -ttl         TTL to use for the tunnel wrapper

          -pidfile     write the PID into this file

Options may be abbreviated with unique prefixes.

EOB
;
  exit;
}



1;
