package IO::Socket::Multicast6;

use strict;
use vars qw(@ISA @EXPORT_OK @EXPORT %EXPORT_TAGS $VERSION);

use IO::Socket;
use Socket6 0.12; 
require IO::Socket::INET6; 
use Net::IP;

use Carp 'croak';
require Exporter;
require DynaLoader;
eval <<END; # or warn "IO::Interface module not installed; Cannot use interface names.\n";
  use IO::Interface 0.94 'IFF_MULTICAST';
END

my @functions = qw(mcast_add mcast_drop mcast_if mcast_loopback 
		   mcast_ttl mcast_dest mcast_send);

%EXPORT_TAGS = ('all'       => \@functions,
		'functions' => \@functions);
@EXPORT = ( );
@EXPORT_OK = @{ $EXPORT_TAGS{'all'} };

@ISA = qw(Exporter DynaLoader IO::Socket::INET6);
$VERSION = '0.10';

sub import {
  Socket->export_to_level(1,@_);
  IO::Socket::Multicast6->export_to_level(1,@_);
}

sub new {
  my $class = shift;
  unshift @_,(Proto => 'udp') unless @_;
  $class->SUPER::new(@_);

}

sub configure {
  my($self,$arg) = @_;
  $arg->{Proto} ||= 'udp';
  $self->SUPER::configure($arg);
}

sub mcast_add {
  my $sock = shift;
  my $group = shift || 
    croak 'usage: $sock->mcast_add($mcast_addr [,$interface])';

  my $n = new Net::IP($group);
  if (not defined $n or $n->version() == 4) {
    croak 'Used inappropriate ip address format IPv4 instead of IPv6'; 
  }

  my $interface = shift; 
  if (not defined $interface) {
    $interface = ""; 
  }

  return $sock->_mcast_add($group,$interface);
}

sub mcast_drop {
  my $sock = shift;
  my $group = shift || 
    croak 'usage: $sock->drop($mcast_addr)';

  my $n = new Net::IP($group);
  if (not defined $n or $n->version() == 4) {
    croak 'Used inappropriate ip address format IPv4 instead of IPv6';
  }

  my $interface = shift;
  if (not defined $interface) {
    $interface = $sock->_mcast_if; 
  }

  return $sock->_mcast_drop($group,$interface);
}

# set the outgoing interface to this interface 
sub mcast_if {
  my $sock = shift;
  my $interface = shift;
  my $previous = $sock->_mcast_if;
  return $sock->_mcast_if($interface) ? $previous : undef;
}

# get/set the destination. 
sub mcast_dest {
  my $sock = shift;
  my $dest = shift; 
  my $prev = ${*$sock}{'io_socket_mcast_dest'};

  if (defined $dest) { 
    croak "invalid destination address" unless length($dest) == 28;
    ${*$sock}{'io_socket_mcast_dest'} = $dest;
  }
  return $prev;
}

sub mcast_send {
  my $sock = shift;
  my $data = shift ||
    croak 'usage: $sock->mcast_send($data [,$address])';
  $sock->mcast_dest(shift) if @_;
  my $dest = $sock->mcast_dest;
  
  my ($port, $host) = unpack_sockaddr_in6($dest); 
  $host = inet_ntop(AF_INET6, $host); 

  return $sock->send($data,0,$dest);
}

bootstrap IO::Socket::Multicast6 $VERSION;

1;
__END__

=head1 NAME

IO::Socket::Multicast6 - Send and receive IPv6 multicast messages

=head1 SYNOPSIS

  use Socket6; 
  use IO::Socket::Multicast6;

  # create a new UDP socket ready to read datagrams on port 1100
  my $s = IO::Socket::Multicast6->new(LocalPort=>1100);

  # Add a multicast group to eth0 device
  $s->mcast_add('FF02:0:0:0:0:0:0:6061','eth0');

  # now receive some multicast data
  $s->recv($data,1024);

  # Drop a multicast group
  $s->mcast_drop('FF02:0:0:0:0:0:0:6061');

  # Set outgoing interface to eth0
  $s->mcast_if('eth0');

  # Set time to live on outgoing multicast packets
  $s->mcast_ttl(10);

  # Turn off loopbacking
  $s->mcast_loopback(0);

  my $port = 1200;
  my $addr = "FF02:0:0:0:0:0:0:6061";
  my $dest = pack_sockaddr_in6($port, inet_pton(AF_INET6, $addr));

  # Multicast a message to group FF02:0:0:0:0:0:0:6061
  $s->mcast_send('hello world!', $dest);

  #alternatively set the destination first and then send data
  $s->mcast_set($dest);
  $s->mcast_send('hello again!');

=head1 DESCRIPTION

The IO::Socket::Multicast6 module subclasses IO::Socket::INET6 to
enable you to manipulate multicast groups.  With this module (and an
operating system that supports multicasting), you will be able to
receive incoming IPv6 multicast transmissions and generate your own
outgoing IPv6 multicast packets.

This module requires IO::Interface version 0.94 or higher.

=head2 INTRODUCTION

Multicasting is designed for streaming multimedia applications and for
conferencing systems in which one transmitting machines needs to
distribute data to a large number of clients.

IPv6 addresses with the prefix 0xFF are reserved for multicasting.
These addresses do not correspond to individual machines, but to
multicast groups.  Messages sent to these addresses will be delivered
to a potentially large number of machines that have registered their
interest in receiving transmissions on these groups.  They work like
TV channels.  A program tunes in to a multicast group to receive
transmissions to it, and tunes out when it no longer wishes to receive
the transmissions.

To receive transmissions B<from> a multicast group, you will use
IO::Socket::Multicast6->new() to create a UDP socket and bind it to a
local network port.  You will then subscribe one or more multicast
groups using the mcast_add() method.  Subsequent calls to the standard
recv() method will now receive messages incoming messages transmitted
to the subscribed groups using the selected port number.

To send transmissions B<to> a multicast group, you can use the
standard send() method to send messages to the multicast group and
port of your choice.  The mcast_set() and mcast_send() methods are
provided as convenience functions.  Mcast_set() will set a default
multicast destination for messages which you then send with
mcast_send().

Note that mcast_send() and mcast_set() in IO::Socket::Multicast6 uses
sockaddr_in6 structure instead of the <IPv4 address>:<port>
combination.

To set the number of hops (routers) that outgoing multicast messages
will cross, call mcast_ttl().  To activate or deactivate the looping
back of multicast messages (in which a copy of the transmitted
messages is received by the local machine), call mcast_loopback().

=head2 CONSTRUCTORS

=over 4

=item $socket = IO::Socket::Multicast6->new(LocalPort=>$port)

The new() method is the constructor for the IO::Socket::Multicast6
class.  It takes the same arguments as IO::Socket::INET6, except that
the B<Proto> argument, rather than defaulting to "tcp", will default
to "udp", which is more appropriate for multicasting.

To create a UDP socket suitable for sending outgoing multicast
messages, call new() without no arguments (or with
C<Proto=E<gt>'udp'>).  To create a UDP socket that can also receive
incoming multicast transmissions on a specific port, call new() with
the B<LocalPort> argument.

=back

=head2 METHODS

=over 4

=item $success = $socket->mcast_add($multicast_address, $interface)

The mcast_add() method will add the provided multicast address to the
list of subscribed multicast groups.  The address and port combination
is presented as a packed IPv6 address (such as produced by the
pack_sockaddr_in6() function).  The $interface argument is used to
specify the network interface for incoming multicast messages. On
success, the method will return a true value.

=item $success = $socket->mcast_drop($multicast_address)

This reverses the action of mcast_add(), removing the indicated
multicast address from the list of subscribed groups.

=item $loopback = $socket->mcast_loopback

=item $previous = $socket->mcast_loopback($new)

The mcast_loopback() method controls whether the socket will receive
its own multicast transmissions (default yes).  Called without
arguments, the method returns the current state of the loopback
flag. Called with a boolean argument, the method will set the loopback
flag, and return its previous value.

=item $ttl = $socket->mcast_ttl

=item $previous = $socket->mcast_ttl($new)

The mcast_ttl() method examines or sets the time to live (TTL) for
outgoing multicast messages.  The TTL controls the numbers of routers
the packet can cross before being expired.  The default TTL is 1,
meaning that the message is confined to the local area network.
Values between 0 and 255 are valid.

Called without arguments, this method returns the socket's current
TTL.  Called with a value, this method sets the TTL and returns its
previous value.

=item $interface = $socket->mcast_if

=item $previous = $socket->mcast_if($new)

By default, the OS will pick the network interface to use for outgoing
multicasts automatically.  You can control this process by using the
mcast_if() method to set the outgoing network interface explicitly.
Called without arguments, returns the current interface.  Called with
the name of an interface, sets the outgoing interface and returns its
previous value.

B<NOTE>: To set the interface used for B<incoming> multicasts, use the
mcast_add() method.

=item $dest = $socket->mcast_dest

=item $previous = $socket->mcast_dest($new)

The mcast_dest() method is a convenience function that allows you to
set the default destination group for outgoing multicasts.  Called
without arguments, returns the current destination as a packed binary
sockaddr_in6 data structure.  Called with a new destination (packed
binary sockaddr_in6 structure), the method sets the default
destination and returns the previous one, if any.

=item $bytes = $socket->mcast_send($data [,$dest])

Mcast_send() is a convenience function that simplifies the sending of
multicast messages.  C<$data> is the message contents, and C<$dest> is
an optional destination group.  The destination must be a packed
sockaddr_in6 structure.  If the destination is not supplied, it will
default to the most recent value set in mcast_dest() or a previous
call to mcast_send().

The method returns the number of bytes successfully queued for
delivery.

As a side-effect, the method will call mcast_dest() to remember the
destination address.

Example:

  my $port = 1500;
  my $addr = "FF02:0:0:0:0:0:0:6061";
  my $dest = pack_sockaddr_in6($port, inet_pton(AF_INET6, $addr));
  $socket->mcast_send('Hi there group members!', $dest) || die;
  $socket->mcast_send("How's the weather?") || die;

=back

=head1 EXAMPLE

The following is an example of a multicast server.  Every 10 seconds
it transmits the current time and the list of logged-in users to the
local network using multicast group FF0F:0:0:0:0:0:0:FF0F, port 2000
(these are chosen arbitrarily).

 #!/usr/bin/perl
 # server
 use strict;

 use Socket6; 
 use IO::Socket::Multicast6;

 use constant GROUP => 'FF0F:0:0:0:0:0:0:FF0F';
 use constant PORT => 2000; 

 my $dest = pack_sockaddr_in6(PORT, inet_pton(AF_INET6, GROUP));
 my $sock = IO::Socket::Multicast6->new(Proto=>'udp'); 
 $sock->mcast_dest($dest); 

 my $interface = "xl0"; 
 $sock->mcast_add(GROUP, $interface); 

 while (1) {
   my $message = localtime;
   $message .= "\n" . `who`;
   $sock->send($message) || die "Couldn't send: $!";
 } continue {
   sleep 10;
 }

This is the corresponding client.  It listens for transmissions on
group FF0F:0:0:0:0:0:0:FF0F, port 2000, and echoes the messages to
standard output.

 #!/usr/bin/perl
 # client

 use strict;
 use Socket6; 
 use IO::Socket::Multicast6;

 use constant GROUP => 'FF0F:0:0:0:0:0:0:FF0F';
 use constant PORT => 2000; 
 my $dest = pack_sockaddr_in6(PORT, inet_pton(AF_INET6, GROUP));
 my $interface = 'xl0';

 my $sock = IO::Socket::Multicast6->new(Proto=>'udp');
 $sock->mcast_add(GROUP, $interface) || die "Couldn't set group: $!\n";

 while (1) {
   my $data;
   my $peer = $sock->recv($data,1024);
   next unless $peer; 
   my ($port, $host) = unpack_sockaddr_in6 ($peer); 
   $host = inet_ntop(AF_INET6, $host); 
   print "Message received from ($port, $host): \n";
   print $data;
 }

=head2 EXPORT

None by default.  However, if you wish to call mcast_add(),
mcast_drop(), mcast_if(), mcast_loopback(), mcast_ttl, mcast_dest()
and mcast_send() as functions you may import them explicitly on the
B<use> line or by importing the tag ":functions".

=head2 BUGS

The mcast_if(), mcast_ttl() and mcast_loopback() methods will cause a
crash on versions of Linux earlier than 2.2.0 because of a kernel bug
in the implementation of the multicast socket options.

The mcast_add() function might fail if the interface is not specified.

=head1 AUTHOR

Lincoln Stein, lstein@cshl.org. (Main, IPv4)
Venkata Pingali, pingali@isi.edu (IPv6 Extension) 

This module is distributed under the same terms as Perl itself.

=head1 SEE ALSO

perl(1), IO::Socket(3), IO::Socket::INET(3), Socket6(3), IO::Socket::INET6(3)

=cut
